"""The nytid courses command and subcommands"""

import typerconf

from nytid import courses
import logging
import os
import pathlib
from nytid.courses import registry
import re
from nytid import storage
import subprocess
import sys
import typer
import typing
from typing_extensions import Annotated

import os
import subprocess

MINE = "mine"

cli = typer.Typer(help="Manage courses")


def complete_register_arg(incomplete):
    """
    Returns all matching register names that can complete `incomplete`.
    """
    return filter(lambda x: x.startswith(incomplete), registry.ls())


def complete_register(ctx, param, incomplete: str):
    """
    Returns list of matching register names available.
    """
    return filter(lambda x: x.startswith(incomplete), registry.ls())


def complete_course_name(ctx: typer.Context, incomplete: str):
    """
    Returns a list of course names matching `incomplete`.
    """
    registers = register_pairs(ctx.params.get("register"))
    for register, register_path in registers:
        courses_in_reg = courses.all_courses(register_path)
        matching_courses = filter(lambda x: x.startswith(incomplete), courses_in_reg)
        return map(lambda x: (x, f"from {register}"), matching_courses)


def register_pairs(register=None):
    """
    Returns a list of (name, path)-tuples (pairs) for registers to use.

    If `register` is None, we use all existing registers found
    in the config. Otherwise, we look up the path of the one specified and return
    a list containing only that name--path-tuple.
    """
    if register:
        return [(register, registry.get(register))]
    else:
        return [(register, registry.get(register)) for register in registry.ls()]


def complete_config_path(ctx: typer.Context, incomplete: str):
    """
    Returns all valid paths in the config starting with `incomplete`.
    """
    register = ctx.params.get("register")
    course = ctx.params.get("course")
    try:
        conf = courses.get_course_config(course, register)
    except:
        return []

    return filter(lambda x: x.startswith(incomplete), conf.paths())


def print_config(conf: typerconf.Config, path: str = ""):
    """
    Prints the config tree contained in `conf` to stdout.
    Optional `path` is prepended.
    """
    try:
        for key in conf.keys():
            if path:
                print_config(conf[key], f"{path}.{key}")
            else:
                print_config(conf[key], key)
    except AttributeError:
        print(f"{path} = {conf}")


def complete_my_course_arg(incomplete: str) -> typing.List[str]:
    """
    Returns a list of courses in my courses that match `incomplete`.
    """
    mine_path = registry.get(MINE)
    for course in courses.all_courses(mine_path):
        if course.startswith(incomplete):
            yield course


def registers_regex(regex: str) -> typing.List[str]:
    """
    Returns a list of registers matching the `regex`.

    Matching is done `re.match`.
    """
    pattern = re.compile(regex)

    registers = registry.ls()
    for register in registers:
        if pattern.match(register):
            yield register


def courses_regex(
    regex: str, registers: typing.List[str] = None
) -> typing.List[typing.Tuple[str, str]]:
    """
    Returns a list of course--register pairs matching the `regex`.

    Matching is done `re.match`.
    """
    pattern = re.compile(regex)

    if not registers:
        registers = registry.ls()

    for register in registers:
        the_courses = courses.all_courses(registry.get(register))
        for course in the_courses:
            if pattern.match(course):
                yield (course, register)


def complete_course_regex(ctx: typer.Context, regex: str) -> typing.List[str]:
    """ """
    register_regex = ctx.params.get("register")
    registers = registers_regex(register_regex)

    return courses_regex(regex, registers)


new_register_arg = typer.Argument(help="A name to refer to the register.")
register_path_arg = typer.Argument(
    help="The absolute path to the register " "directory."
)
register_arg = typer.Argument(
    help="The name of the register.", autocompletion=complete_register_arg
)
alias_opt = typer.Option(help="Alias to use instead of " "the original course name")
my_course_arg = typer.Argument(
    help="The course in my courses.", autocompletion=complete_my_course_arg
)
register_opt_regex = typer.Option(
    help="Regex for register names to use.", shell_complete=registers_regex
)
course_arg_regex = typer.Argument(
    help="Regex matching courses.", autocompletion=complete_course_regex
)
registrycli = typer.Typer(name="registry", help="Manage course registers")

cli.add_typer(registrycli)


@registrycli.command(name="ls")
def registry_ls():
    """
    Lists registers added to the configuration.
    """
    for register in registry.ls():
        print(f"{register}\t{registry.get(register)}")


@registrycli.command(name="add")
def registry_add(
    name: Annotated[str, new_register_arg],
    register_path: Annotated[pathlib.Path, register_path_arg],
):
    """
    Adds a register to the configuration.
    """
    try:
        registry.add(name, register_path)
    except KeyError as err:
        logging.error(f"Can't add {name}: {err}")
        sys.exit(1)


@registrycli.command(name="rm")
def registry_rm(name: Annotated[str, register_arg]):
    """
    Removes a register from the configuration.
    """
    try:
        registry.remove(name)
    except KeyError as err:
        logging.error(f"Can't remove {name}: {err}")
        sys.exit(1)


course_arg = typer.Argument(help="A name used to refer to the course.")
register_option = typer.Option(
    help="Name of register to use. "
    "Must be used if there are more than "
    "one register in the config.",
    shell_complete=complete_register,
)
contact_opt = typer.Option(
    help="The course responsible's contact info. "
    "Default can be set using "
    "`nytid config me.name --set 'First Last'"
    " and "
    "`nytid config me.email --set x@y.z."
)
code_opt = typer.Option(
    help="The course code, " "to relate it to similar courses using " "`related`."
)
related_opt = typer.Option(
    help="List of related course codes, "
    "courses with any of these codes can "
    "share TAs."
)
ics_opt = typer.Option(
    help="A URL to an ICS file containing the "
    "schedule. E.g. an export/subscription from "
    "TimeEdit."
)
data_path_opt = typer.Option(
    help="Path to the course's data directory, "
    "default is to append `/data` to the "
    "course's config directory."
)
num_students_opt = typer.Option(help="The total number of students " "in the course.")
num_groups_opt = typer.Option(
    help="The number of groups that the class " "will be divided into."
)


@cli.command()
def new(
    name: Annotated[str, course_arg],
    register: Annotated[str, register_option] = MINE,
    contact: Annotated[str, contact_opt] = None,
    code: Annotated[str, code_opt] = None,
    related: Annotated[typing.List[str], related_opt] = [],
    ics: Annotated[str, ics_opt] = None,
    data_path: Annotated[pathlib.Path, data_path_opt] = None,
    num_students: Annotated[int, num_students_opt] = None,
    num_groups: Annotated[int, num_groups_opt] = None,
):
    """
    Creates a new course.
    """
    kwdata = {
        "contact": contact,
        "code": code,
        "related_codes": related,
        "ics": ics,
        "data_path": data_path,
        "num_students": num_students,
        "num_groups": num_groups,
    }
    courses.new(name, register, kwdata)


course_arg_autocomplete = typer.Argument(
    help="Name (nickname) of the target " "course.", autocompletion=complete_course_name
)
path_arg = typer.Argument(
    help="Path in config, e.g. 'courses.datintro22'. "
    "Empty string is root of config. Defaults to "
    "the empty string.",
    autocompletion=complete_config_path,
)
value_arg = typer.Option(
    "-s",
    "--set",
    help="Values to store. "
    "More than one value makes a list. "
    "Values are treated as JSON if possible.",
)


@cli.command(name="config")
def config_cmd(
    course: Annotated[str, course_arg_autocomplete],
    register: Annotated[str, register_option] = MINE,
    path: Annotated[str, path_arg] = "",
    values: Annotated[typing.List[str], value_arg] = None,
):
    """
    Reads values from or writes `values` to the config of `course` at `path`.
    """
    try:
        conf = courses.get_course_config(course, register)
    except KeyError as err:
        logging.error(err)
        sys.exit(1)
    if values:
        if len(values) == 1:
            values = values[0]
        if values == "":
            values = None
        conf.set(path, values)
    else:
        try:
            print_config(conf.get(path), path)
        except KeyError as err:
            logging.error(f"{path} doen't exist in the config: {err}")
            sys.exit(1)


@cli.command()
def ls(register: Annotated[str, register_option] = MINE):
    """
    Lists all available courses in all registers in the registry. Output format:

      register<tab>course

    If `register` (a register name) is provided, only courses from that register
    are listed.
    """
    if register:
        try:
            for course in courses.all_courses(registry.get(register)):
                print(f"{register}\t{course}")
        except KeyError as err:
            logging.error(err)
            sys.exit(1)
    else:
        for register in registry.ls():
            for course in courses.all_courses(registry.get(register)):
                print(f"{register}\t{course}")


datacli = typer.Typer(name="data", help="Access the raw course data")

cli.add_typer(datacli)


@datacli.command()
def shell(
    course: Annotated[str, course_arg_autocomplete],
    register: Annotated[str, register_arg] = MINE,
):
    """
    Spawns a shell in the data directory of a course.
    """
    try:
        data_dir = courses.get_course_data(course, register)
    except KeyError as err:
        logging.error(err)
        sys.exit(1)
    try:
        print(f"--- {course}/data shell ---")
        env = os.environ.copy()
        if "PS1" in env:
            env["PS1"] = f"{course}/data {env['PS1']}"
        else:
            env["PS1"] = f"{course}/data \w\n$ "

        subprocess.run(
            [os.environ["SHELL"]],
            cwd=data_dir.path,
            env=env,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
    except FileNotFoundError as err:
        logging.warning(f"The data directory doesn't exist: {err}")
        try:
            os.makedirs(data_dir.path)
            logging.warning(f"Created {data_dir.path}")
            env = os.environ.copy()
            if "PS1" in env:
                env["PS1"] = f"{course}/data {env['PS1']}"
            else:
                env["PS1"] = f"{course}/data \w\n$ "

            subprocess.run(
                [os.environ["SHELL"]],
                cwd=data_dir.path,
                env=env,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        except Exception as err:
            logging.error(f"Can't create directory: {err}")
            sys.exit(1)
    finally:
        print(f"--- {course}/data shell terminated ---")


minecli = typer.Typer(name=MINE, help="Manage my courses")

cli.add_typer(minecli)


@minecli.command(name="add")
def mine_add(
    course: Annotated[str, course_arg_autocomplete],
    register: Annotated[str, register_arg] = None,
    alias: Annotated[str, alias_opt] = None,
):
    """
    Adds a course to my courses.
    """
    if not register:
        register_map = {register: registry.get(register) for register in registry.ls()}

        matching_registers = []
        for register, register_path in register_map.items():
            matching_courses = filter(
                lambda x: x == course, courses.all_courses(register_path)
            )
            if len(list(matching_courses)) > 0:
                matching_registers.append(register)

        num_matches = len(matching_registers)
        if num_matches == 1:
            register = matching_registers[0]
        elif num_matches < 1:
            logging.error(f"Can't find course {course} in any register")
            sys.exit(1)
        elif num_matches > 1:
            logging.error(
                f"Too many matches for {course}, "
                f"specify one of the registers {matching_registers}."
            )
            sys.exit(1)

    if not alias:
        alias = course

    try:
        mine_path = registry.get(MINE)
    except KeyError:
        home_path = pathlib.Path(os.environ["HOME"])
        registry.add(MINE, home_path / ".nytid/mine")
        mine_path = registry.get(MINE)
        os.makedirs(mine_path)
        logging.warning(f"Added register {MINE} and created its path {mine_path}.")
        mine_path = registry.get(MINE)
    for my_course in mine_path.iterdir():
        if my_course == alias:
            logging.error(f"{alias} is already among your courses.")
            sys.exit(1)

    course_path = registry.get(register) / course
    os.symlink(course_path, mine_path / alias)


@minecli.command(name="ls")
def mine_ls():
    """
    Lists my courses.
    """
    mine_path = registry.get(MINE)
    for course in mine_path.iterdir():
        print(course.name)


@minecli.command(name="rm")
def mine_rm(course: Annotated[str, my_course_arg]):
    """
    Removes a course from my courses.
    """
    mine_path = registry.get(MINE)
    try:
        os.unlink(mine_path / course)
    except FileNotFoundError as err:
        logging.error(f"Couldn't find {course}: {err}")
        sys.exit(1)


if __name__ == "__main__":
    cli()



import arrow
import datetime
from nytid.signup import sheets
from nytid.signup.sheets import SIGNUP_SHEET_HEADER


def time_for_event(event, amanuensis=False):
    """
    Input: an event of type ics.event.Event and an optional bool amanuensis
    specifying whether the computation is for an amanuensis or not.

    Output: Returns a datetime.timedelta corresponding to the time including prep
    time for the event.
    """
    return add_prep_time(
        round_time(event.end - event.begin), event.name, event.begin, amanuensis
    )


def round_time(time):
    """
    Input: A datetime.timedelta object time.

    Output: The time object rounded according to KTH rules. Currently round up to
    nearest quarter of an hour.
    """
    HOUR = 60 * 60
    QUARTER = 15 * 60

    total_seconds = time.total_seconds()
    full_hours = (total_seconds // HOUR) * HOUR
    part_hour = total_seconds % HOUR

    if part_hour > 3 * QUARTER:
        return datetime.timedelta(seconds=full_hours + HOUR)
    elif part_hour > 2 * QUARTER:
        return datetime.timedelta(seconds=full_hours + 3 * QUARTER)
    elif part_hour > QUARTER:
        return datetime.timedelta(seconds=full_hours + 2 * QUARTER)
    elif part_hour > 0:
        return datetime.timedelta(seconds=full_hours + QUARTER)

    return datetime.timedelta(seconds=full_hours)


def add_prep_time(time, event_type, date=datetime.date.today(), amanuensis=False):
    """
    Input:
    - a datetime.timedelta object time,
    - a string containing the title of the event,
    - an optional date (datetime or arrow object) indicating the date of the
      event. If no date is given, today's date is assumed, meaning the latest
      prep-time policy will be used.
    - an optional bool indicating amanuensis employment or hourly.

    Output: the time object rounded according to KTH rules.
    """
    time *= prep_factor(event_type, date=date, amanuensis=amanuensis)
    return round_time(time)


def prep_factor(event_type, date, amanuensis):
    """ """
    if isinstance(date, arrow.Arrow):
        date = date.datetime
    if isinstance(date, datetime.datetime):
        date = date.date()

    event_type = event_type.casefold()

    if "övning" in event_type or "exercise" in event_type:
        return 2

    tutoring = ["laboration", "seminar", "redovisning", "presentation"]

    if date < datetime.date(2023, 1, 1) or (
        date < datetime.date(2022, 10, 1) and not amanuensis
    ):
        if check_substrings(tutoring, event_type):
            return 1.33
    else:
        if check_substrings(tutoring, event_type):
            if "online" in event_type:
                return 1.5
            return 1.8

    return 1


def check_substrings(substrings, string):
    """
    Check if any of the substrings (list) is a substring of string.
    Return bool.
    """
    for substring in substrings:
        if substring in string:
            return True

    return False


def hours_per_student(csv_rows, round_time=round_time):
    """
    Input: The schedule as rows of CSV data as from csv.reader.

    Output: A dictionary mapping event type to the total number of hours per
    student.
    """

    start_index = SIGNUP_SHEET_HEADER.index("Start")
    end_index = SIGNUP_SHEET_HEADER.index("End")
    event_index = SIGNUP_SHEET_HEADER.index("Event")

    event_hours = dict()

    for row in csv_rows:
        time = round_time(
            arrow.get(row[end_index], "YYYY-MM-DD HH:mm")
            - arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
        )
        event_type = row[event_index]

        if event_type not in event_hours:
            event_hours[event_type] = time
        else:
            event_hours[event_type] += time

    return event_hours


def hours_per_TA(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
    """
    Input:
    - Rows of CSV data as from csv.reader.
    - add_prep_time allows using a different function for adding prep time than
      the default.
    - round_time allows using a different function for rounding that the default.

    Output: a dictionary mapping a TA to the number of hours they signed up for
    (not counting slots where they're in reserve position) in the sign-up sheet,
    {TA: hours}. The hours as datetime.timedelta objects.
    """
    TA_hours = {}

    start_index = SIGNUP_SHEET_HEADER.index("Start")
    end_index = SIGNUP_SHEET_HEADER.index("End")
    event_index = SIGNUP_SHEET_HEADER.index("Event")

    for row in csv_rows:
        start_date = arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
        time = arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - start_date

        time = round_time(time)
        time = add_prep_time(time, row[event_index], start_date)

        booked, _ = sheets.get_booked_TAs_from_csv(row)

        for assistant in booked:
            if assistant in TA_hours:
                TA_hours[assistant] += time
            else:
                TA_hours[assistant] = time

    return TA_hours


def total_hours(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
    """
    Input:
    - Rows of CSV data as from csv.reader.
    - add_prep_time allows using a different function for adding prep time than
      the default.
    - round_time allows using a different function for rounding that the default.

    Output: Total number of hours spent on the course, as a datetime.timedelta
    object.
    """
    total = datetime.timedelta(0)
    TA_hours = hours_per_TA(csv_rows, add_prep_time, round_time)

    for _, hours in TA_hours.items():
        total += hours

    return total


def max_hours(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
    """
    Input:
    - takes the rows of CSV as output from csv.reader.
    - add_prep_time allows using a different function for adding prep time than
      the default.
    - round_time allows using a different function for rounding that the default.

    Output: returns the maximum number of hours (using maximum TAs needed), as a
    detetime.timedelta object.
    """
    start_index = SIGNUP_SHEET_HEADER.index("Start")
    end_index = SIGNUP_SHEET_HEADER.index("End")
    event_index = SIGNUP_SHEET_HEADER.index("Event")
    needed_TAs_index = SIGNUP_SHEET_HEADER.index("#Needed TAs")

    max_hours = datetime.timedelta()

    for row in csv_rows:
        start_date = arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
        time = arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - start_date

        time = round_time(time)
        time = add_prep_time(time, row[event_index], start_date)

        max_num_TAs = int(row[needed_TAs_index])

        max_hours += time * max_num_TAs

    return max_hours


def hours_per_event(csv_rows, round_time=round_time):
    """
    Input:
    - Rows of CSV data as from csv.reader.
    - round_time allows using a different function for rounding that the default.

    Output: a dictionary mapping an event type to the number of hours assigned to
    that type of event in the sign-up sheet, {event: hours}. The hours as
    datetime.timedelta objects.
    """
    event_hours = {}

    start_index = SIGNUP_SHEET_HEADER.index("Start")
    end_index = SIGNUP_SHEET_HEADER.index("End")
    event_index = SIGNUP_SHEET_HEADER.index("Event")

    for row in csv_rows:
        time = round_time(
            arrow.get(row[end_index], "YYYY-MM-DD HH:mm")
            - arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
        )

        if row[event_index] in event_hours:
            event_hours[row[event_index]] += time
        else:
            event_hours[row[event_index]] = time

    return event_hours


def compute_amanuensis_data(
    csv_rows,
    low_percentage=0.04,
    min_days=25,
    add_prep_time=add_prep_time,
    round_time=round_time,
    begin_date=None,
):
    """
    Input:
    - csv_rows, the CSV rows as output from csv.reader.
    - low_percentage, the lowest acceptable percentage of an amanuensis
      contract.
    - min_days is the minimum number of days we accept for an amanuensis
      contract.
    - add_prep_time allows using a different function for adding prep time than
      the default.
    - round_time allows using a different function for rounding that the default.
    - begin_date means that we will force this date as a start date, None means
      we will compute the start date.

    Output: a dictionary {TA: (start, end, hours)} mapping the TA username to a
    tuple (start, end, hours) with the start and end time and the total number of
    hours.
    """
    start_index = SIGNUP_SHEET_HEADER.index("Start")
    end_index = SIGNUP_SHEET_HEADER.index("End")
    event_index = SIGNUP_SHEET_HEADER.index("Event")
    needed_TAs_index = SIGNUP_SHEET_HEADER.index("#Needed TAs")

    ta_hours = hours_per_TA(csv_rows, add_prep_time, round_time)
    ta_data = {}

    for ta in ta_hours.keys():
        earliest_date = arrow.get(csv_rows[0][start_index], "YYYY-MM-DD")
        latest_date = arrow.get(csv_rows[0][end_index], "YYYY-MM-DD")

        for row in csv_rows:
            start_date = arrow.get(row[start_index], "YYYY-MM-DD")
            end_date = arrow.get(row[end_index], "YYYY-MM-DD")

            if start_date < earliest_date:
                earliest_date = start_date
            if end_date > latest_date:
                latest_date = end_date

        hours = ta_hours[ta].total_seconds() / 60 / 60

        july = arrow.get(earliest_date.year, 7, 1)

        if earliest_date < july:
            semester_start = arrow.get(earliest_date.year, 1, 1)
            semester_end = arrow.get(earliest_date.year, 6, 30)
        else:
            semester_start = arrow.get(earliest_date.year, 8, 1)
            semester_end = arrow.get(earliest_date.year + 1, 1, 31)

        earliest_month = arrow.get(earliest_date.year, earliest_date.month, 1)
        latest_month = arrow.get(latest_date.year, latest_date.month, 1).shift(
            months=1, seconds=-1
        )

        if begin_date:
            semester_start = earliest_date = earliest_month = begin_date

        if latest_date > semester_end:
            semester_end = latest_month

        if compute_percentage(semester_start, semester_end, hours) >= low_percentage:
            earliest_date = semester_start
            latest_date = semester_end
        elif compute_percentage(semester_start, latest_month, hours) >= low_percentage:
            earliest_date = earliest_month
            latest_date = latest_month
        elif compute_percentage(earliest_month, semester_end, hours) >= low_percentage:
            earliest_date = earliest_month
            latest_date = latest_month
        elif compute_percentage(earliest_month, latest_month, hours) >= low_percentage:
            earliest_date = earliest_month
            latest_date = latest_month
        elif compute_percentage(earliest_date, latest_month, hours) >= low_percentage:
            latest_date = latest_month
        elif compute_percentage(earliest_date, latest_date, hours) >= low_percentage:
            pass
        else:
            continue  # skip to next TA

        if latest_date - earliest_date < datetime.timedelta(days=min_days):
            continue

        ta_data[ta] = (earliest_date, latest_date, hours)

    return ta_data


def compute_percentage(start, end, hours):
    """
    Input: start and end as arrow.arrow.Arrow or datetime.date objects,
      hours as a float.

    Output: a float in the interval [0, 1], which is the percentage of full time.
    """
    days = (end - start).total_seconds() / 60 / 60 / 24
    return (hours / 1600) * (365 / days)



import datetime
from enum import Enum
import ics.icalendar
import logging
import pathlib
import sys
import typer
from typing_extensions import Annotated

from nytid.cli import courses as coursescli
from nytid.cli.signupsheets import SIGNUPSHEET_URL_PATH
from nytid import courses as courseutils
from nytid import schedules as schedutils
from nytid.signup import hr
from nytid.signup import sheets

import os
from nytid.signup import sheets
from nytid.signup import utils
from nytid.signup import hr

cli = typer.Typer(name="hr", help="Manage sign-up sheets for teaching")


def to_hours(td):
    return td.total_seconds() / 60 / 60


try:
    default_username = os.environ["USER"]
except KeyError:
    default_username = None

username_opt = typer.Option(
    help="Username to filter sign-up sheet for, "
    "defaults to logged in user's username."
)


@cli.command()
def time(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Summarizes the time spent on teaching the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    h_per_student = hr.hours_per_student(booked)

    for event, hours in h_per_student.items():
        print(f"{event}: {to_hours(hours):.2f} h/student")

    print(
        f"Booked: {to_hours(hr.total_hours(booked)):.2f} h "
        f"({to_hours(hr.max_hours(booked)):.2f} h)\n"
    )

    print("# Amanuenser")

    amanuensis = hr.compute_amanuensis_data(booked)

    for user, data in amanuensis.items():
        if not user:
            continue
        print(
            f"{user}: {data[2]:.2f} h, "
            f"{100*hr.compute_percentage(*data):.1f}%: "
            f"{data[0].format('YYYY-MM-DD')}--{data[1].format('YYYY-MM-DD')}"
        )

    print()
    print("# Hourly")

    for user, hours in hr.hours_per_TA(booked).items():
        if not user or user in amanuensis:
            continue
        print(f"{user}: {to_hours(hours):.2f} h")


@cli.command()
def users(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Prints the list of all usernames booked on the course.
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    for user in hr.hours_per_TA(booked):
        print(user)



import arrow
import csv
import datetime
import ics.attendee
import ics.event
import nytid.schedules
import re
import requests

SIGNUP_SHEET_HEADER = [
  "Event", "Start", "End", "#Rooms",
  "#Needed TAs"
]
class EventFromCSV(ics.event.Event):
  """A class to create an ics.event.Event from an event in CSV format"""
  def __init__(self, csv_row):
    """
    Input: a row from a calendar in CSV format (e.g. the sign-up sheet).
    """
    attribute_map = {
      SIGNUP_SHEET_HEADER.index("Event"): "name",
      SIGNUP_SHEET_HEADER.index("Start"): "begin",
      SIGNUP_SHEET_HEADER.index("End"): "end",
      SIGNUP_SHEET_HEADER.index("#Rooms"): "description",
      SIGNUP_SHEET_HEADER.index("#Needed TAs"): "description"
    }
    for idx in range(len(SIGNUP_SHEET_HEADER), len(csv_row)):
      attribute_map[idx] = "attendees"

    kwargs = dict()

    for column, attribute in attribute_map.items():
      try:
        value = csv_row[column]

        if attribute == "description":
          if attribute in kwargs:
            value = kwargs[attribute] + "\n" + value
        elif attribute == "attendees":
          if not value:
            continue

          value = ics.attendee.Attendee(f"{value}@kth.se")

          if attribute not in kwargs:
            value = [value]
          else:
            value = kwargs[attribute] + [value]
        elif attribute in ["begin", "end"]:
          value = arrow.get(value, tzinfo="local")

        kwargs[attribute] = value
      except AttributeError:
        pass

    super().__init__(**kwargs)
def needed_TAs(event):
  """
  Takes an event and returns the number of TAs needed
  """
  num_groups = event.description.split().count("grupp")
  if num_groups == 0:
    num_groups = event.description.split().count("group")

  num_rooms = len(event.location.split(","))

  num_TAs = max(num_rooms, num_groups)

  if "laboration" in event.name or "Laboration" in event.name:
    num_TAs = round(num_TAs * 1.5)

  return num_TAs

def event_filter(events):
  """
  Takes a list of events, returns a filtered list of events (generator).
  The events to include are the teaching events.
  """
  events_whitelist = [
    "Datorlaboration", "Laboration",
    "Övning",
    "Seminarium", "Redovisning", "Examination",
    "Föreläsning"
  ]

  for event in events:
    for event_type in events_whitelist:
      if event_type in event.name:
        yield event
        break
  
def generate_signup_sheet(course, url,
  needed_TAs=needed_TAs, event_filter=event_filter):
  """
  Input:
  - course is a string containing the file name used for output.
  - url is the URL to the ICS-formatted calendar.
  - needed_TAs is a function computing the number of needed TAs based on the 
    event. The default is the needed_TAs function in this module,
  - event_filter is a function that filters events, takes a list of events as 
    argument and returns a filtered list.

  Output:
  Returns nothing. Writes output to {course}.csv.
  """
  with open(f"{course}.csv", "w") as out:
    csvout = csv.writer(out)
    calendar = nytid.schedules.read_calendar(url)

    max_num_TAs = 0
    rows = []

    for event in event_filter(calendar.timeline):
      num_TAs = needed_TAs(event)

      if num_TAs > max_num_TAs:
        max_num_TAs = num_TAs

      rows.append([
        event.name,
        event.begin.to("local").format("YYYY-MM-DD HH:mm"),
        event.end.to("local").format("YYYY-MM-DD HH:mm"),
        len(event.location.split(",")),
        num_TAs
      ])

    csvout.writerow(SIGNUP_SHEET_HEADER +
      [f"TA username" for n in range(max_num_TAs)] +
        ["..."])

    csvout.writerows(rows)
def read_signup_sheet_from_file(filename):
  """
  Input: filename is a string containing the file name of the CSV file of the 
  sign-up sheet.

  Output: All the rows of the CSV as a Python list.
  """
  with open(filename, "r") as f:
    csvfile = csv.reader(f)
    return list(filter(any, list(csvfile)[1:]))
def read_signup_sheet_from_url(url):
  """
  Input: url is a string containing the URL of the CSV file of the sign-up 
  sheet.

  Output: All the rows of the CSV as a Python list.
  """
  response = requests.get(url)
  if response.status_code != 200:
    raise ValueError(response.text)

  response.encoding = response.apparent_encoding
  csvdata = response.text.splitlines()
  return list(filter(any, list(csv.reader(csvdata))[1:]))
def google_sheet_to_csv_url(share_url):
  """
  Input: The share URL of a Google Sheets sheet.

  Output: A URL that downloads (exports) the sheet in CSV format.
  """
  match = re.search("/edit.*$", share_url)
  if not match:
    raise ValueError(f"{share_url} doesn't seem like a Google Sheets URL.")

  url = share_url[:match.start()]
  return url + "/export?format=csv"
def get_TAs_from_csv(csv_row):
  """
  Input: takes a CSV data row as from a csv.reader.

  Output: returns the list of signed TAs. Ensures casefold for TA IDs.
  """
  return list(
    map(lambda x: x.casefold(),
      filter(lambda x: x.strip(),
        csv_row[len(SIGNUP_SHEET_HEADER):]))
    )
def get_booked_TAs_from_csv(csv_row):
  """
  Input: takes a CSV data row as from a csv.reader.

  Output: returns the list of signed TAs, the first N, where N is the number of 
  needed TAs specified in the CSV data.
  """
  TAs = get_TAs_from_csv(csv_row)
  num_needed_TAs = int(csv_row[SIGNUP_SHEET_HEADER.index("#Needed TAs")])

  return TAs[:num_needed_TAs], TAs[num_needed_TAs:]
def filter_events_by_TA(ta_id, csv_rows):
  """
  Input: ta_id is the string to (exactly) match the TAs' identifiers against;
  csv_rows is a list of CSV rows, as from csv.reader.

  Output: a list of CSV rows containing only the rows containing ta_id.
  """
  return list(
    filter(lambda x: ta_id.casefold() in get_TAs_from_csv(x),
      csv_rows))
def filter_events_by_title(event_title, csv_rows):
  """
  Input: event_title is the substring to match the event title against;
  csv_rows is a list of CSV rows, as from csv.reader.

  Output: a list of CSV rows containing only the rows with an event title 
  having event_title as substring.
  """
  return list(filter(
    lambda x: event_title in x[SIGNUP_SHEET_HEADER.index("Event")],
    csv_rows))
def time_for_event(event, amanuensis=False):
  """
  Input: an event of type ics.event.Event and an optional bool amanuensis 
  specifying whether the computation is for an amanuensis or not.

  Output: Returns a datetime.timedelta corresponding to the time including prep 
  time for the event.
  """
  return add_prep_time(
    round_time(event.end-event.begin), event.name, event.begin, amanuensis)
def round_time(time):
  """
  Input: A datetime.timedelta object time.

  Output: The time object rounded according to KTH rules. Currently round up to 
  nearest quarter of an hour.
  """
  HOUR = 60*60
  QUARTER = 15*60

  total_seconds = time.total_seconds()
  full_hours = (total_seconds // HOUR) * HOUR
  part_hour = total_seconds % HOUR

  if part_hour > 3*QUARTER:
    return datetime.timedelta(seconds=full_hours+HOUR)
  elif part_hour > 2*QUARTER:
    return datetime.timedelta(seconds=full_hours+3*QUARTER)
  elif part_hour > QUARTER:
    return datetime.timedelta(seconds=full_hours+2*QUARTER)
  elif part_hour > 0:
    return datetime.timedelta(seconds=full_hours+QUARTER)

  return datetime.timedelta(seconds=full_hours)
def add_prep_time(time, event_type,
  date=datetime.date.today(), amanuensis=False):
  """
  Input:
  - a datetime.timedelta object time,
  - a string containing the title of the event,
  - an optional date (datetime or arrow object) indicating the date of the 
    event. If no date is given, today's date is assumed, meaning the latest 
    prep-time policy will be used.
  - an optional bool indicating amanuensis employment or hourly.

  Output: the time object rounded according to KTH rules.
  """
  if isinstance(date, arrow.Arrow):
    date = date.datetime
  if isinstance(date, datetime.datetime):
    date = date.date()

  if date > datetime.date(2023, 1, 1) or \
    (date > datetime.date(2022, 10, 1) and not amanuensis):
    if check_substrings(
          ["laboration", "seminarium", "redovisning"],
          event_type.casefold()):
      time *= 1.5
    elif "övning" in event_type.casefold():
      time *= 2
  else:
    if check_substrings(
          ["laboration", "seminarium", "redovisning"],
          event_type.casefold()):
      time *= 1.33
    elif "övning" in event_type.casefold():
      time *= 2

  return round_time(time)

def check_substrings(substrings, string):
  """
  Check if any of the substrings (list) is a substring of string.
  Return bool.
  """
  for substring in substrings:
    if substring in string:
      return True

  return False
def hours_per_student(csv_rows, round_time=round_time):
  """
  Input: The schedule as rows of CSV data as from csv.reader.

  Output: A dictionary mapping event type to the total number of hours per 
  student.
  """

  start_index = SIGNUP_SHEET_HEADER.index("Start")
  end_index = SIGNUP_SHEET_HEADER.index("End")
  event_index = SIGNUP_SHEET_HEADER.index("Event")

  event_hours = dict()

  for row in csv_rows:
    time = round_time(arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - \
      arrow.get(row[start_index], "YYYY-MM-DD HH:mm"))
    event_type = row[event_index]

    if event_type not in event_hours:
      event_hours[event_type] = time
    else:
      event_hours[event_type] += time

  return event_hours
def hours_per_TA(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
  """
  Input:
  - Rows of CSV data as from csv.reader.
  - add_prep_time allows using a different function for adding prep time than 
    the default.
  - round_time allows using a different function for rounding that the default.

  Output: a dictionary mapping a TA to the number of hours they signed up for 
  (not counting slots where they're in reserve position) in the sign-up sheet, 
  {TA: hours}. The hours as datetime.timedelta objects.
  """
  TA_hours = {}

  start_index = SIGNUP_SHEET_HEADER.index("Start")
  end_index = SIGNUP_SHEET_HEADER.index("End")
  event_index = SIGNUP_SHEET_HEADER.index("Event")

  for row in csv_rows:
    start_date = arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
    time = arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - start_date

    time = round_time(time)
    time = add_prep_time(time, row[event_index], start_date)

    booked, _ = get_booked_TAs_from_csv(row)

    for assistant in booked:
      if assistant in TA_hours:
        TA_hours[assistant] += time
      else:
        TA_hours[assistant] = time

  return TA_hours
def total_hours(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
  """
  Input:
  - Rows of CSV data as from csv.reader.
  - add_prep_time allows using a different function for adding prep time than 
    the default.
  - round_time allows using a different function for rounding that the default.

  Output: Total number of hours spent on the course, as a datetime.timedelta 
  object.
  """
  total = datetime.timedelta(0)
  TA_hours = hours_per_TA(csv_rows, add_prep_time, round_time)

  for _, hours in TA_hours.items():
    total += hours

  return total
def max_hours(csv_rows, add_prep_time=add_prep_time, round_time=round_time):
  """
  Input:
  - takes the rows of CSV as output from csv.reader.
  - add_prep_time allows using a different function for adding prep time than 
    the default.
  - round_time allows using a different function for rounding that the default.

  Output: returns the maximum number of hours (using maximum TAs needed), as a 
  detetime.timedelta object.
  """
  start_index = SIGNUP_SHEET_HEADER.index("Start")
  end_index = SIGNUP_SHEET_HEADER.index("End")
  event_index = SIGNUP_SHEET_HEADER.index("Event")
  needed_TAs_index = SIGNUP_SHEET_HEADER.index("#Needed TAs")

  max_hours = datetime.timedelta()

  for row in csv_rows:
    start_date = arrow.get(row[start_index], "YYYY-MM-DD HH:mm")
    time = arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - start_date

    time = round_time(time)
    time = add_prep_time(time, row[event_index], start_date)

    max_num_TAs = int(row[needed_TAs_index])

    max_hours += time * max_num_TAs

  return max_hours
def hours_per_event(csv_rows, round_time=round_time):
  """
  Input:
  - Rows of CSV data as from csv.reader.
  - round_time allows using a different function for rounding that the default.

  Output: a dictionary mapping an event type to the number of hours assigned to 
  that type of event in the sign-up sheet, {event: hours}. The hours as 
  datetime.timedelta objects.
  """
  event_hours = {}

  start_index = SIGNUP_SHEET_HEADER.index("Start")
  end_index = SIGNUP_SHEET_HEADER.index("End")
  event_index = SIGNUP_SHEET_HEADER.index("Event")

  for row in csv_rows:
    time = round_time(arrow.get(row[end_index], "YYYY-MM-DD HH:mm") - \
      arrow.get(row[start_index], "YYYY-MM-DD HH:mm"))

    if row[event_index] in event_hours:
      event_hours[row[event_index]] += time
    else:
      event_hours[row[event_index]] = time

  return event_hours
def compute_amanuensis_data(csv_rows, low_percentage=0.04, min_days=25,
    add_prep_time=add_prep_time, round_time=round_time, begin_date=None):
  """
  Input:
  - csv_rows, the CSV rows as output from csv.reader.
  - low_percentage, the lowest acceptable percentage of an amanuensis 
    contract.
  - min_days is the minimum number of days we accept for an amanuensis 
    contract.
  - add_prep_time allows using a different function for adding prep time than 
    the default.
  - round_time allows using a different function for rounding that the default.
  - begin_date means that we will force this date as a start date, None means 
    we will compute the start date.

  Output: a dictionary {TA: (start, end, hours)} mapping the TA username to a 
  tuple (start, end, hours) with the start and end time and the total number of 
  hours.
  """
  start_index = SIGNUP_SHEET_HEADER.index("Start")
  end_index = SIGNUP_SHEET_HEADER.index("End")
  event_index = SIGNUP_SHEET_HEADER.index("Event")
  needed_TAs_index = SIGNUP_SHEET_HEADER.index("#Needed TAs")

  ta_hours = hours_per_TA(csv_rows, add_prep_time, round_time)
  ta_data = {}

  for ta in ta_hours.keys():
    earliest_date = arrow.get(csv_rows[0][start_index], "YYYY-MM-DD")
    latest_date = arrow.get(csv_rows[0][end_index], "YYYY-MM-DD")

    for row in csv_rows:
      start_date = arrow.get(row[start_index], "YYYY-MM-DD")
      end_date = arrow.get(row[end_index], "YYYY-MM-DD")

      if start_date < earliest_date:
        earliest_date = start_date
      if end_date > latest_date:
        latest_date = end_date

    hours = ta_hours[ta].total_seconds()/60/60

    july = arrow.get(earliest_date.year, 7, 1)

    if earliest_date < july:
      semester_start = arrow.get(earliest_date.year, 1, 1)
      semester_end = arrow.get(earliest_date.year, 6, 30)
    else:
      semester_start = arrow.get(earliest_date.year, 8, 1)
      semester_end = arrow.get(earliest_date.year+1, 1, 31)

    earliest_month = arrow.get(earliest_date.year, earliest_date.month, 1)
    latest_month = arrow.get(
      latest_date.year, latest_date.month, 1).shift(months=1, seconds=-1)

    if begin_date:
      semester_start = earliest_date = earliest_month = begin_date

    if latest_date > semester_end:
      semester_end = latest_month

    if compute_percentage(semester_start, semester_end, hours) >= low_percentage:
      earliest_date = semester_start
      latest_date = semester_end
    elif compute_percentage(semester_start, latest_month, hours) >= low_percentage:
      earliest_date = earliest_month
      latest_date = latest_month
    elif compute_percentage(earliest_month, semester_end, hours) >= low_percentage:
      earliest_date = earliest_month
      latest_date = latest_month
    elif compute_percentage(earliest_month, latest_month, hours) >= low_percentage:
      earliest_date = earliest_month
      latest_date = latest_month
    elif compute_percentage(earliest_date, latest_month, hours) >= low_percentage:
      latest_date = latest_month
    elif compute_percentage(earliest_date, latest_date, hours) >= low_percentage:
      pass
    else:
      continue # skip to next TA

    if latest_date - earliest_date < datetime.timedelta(days=min_days):
      continue

    ta_data[ta] = (earliest_date, latest_date, hours)

  return ta_data
def compute_percentage(start, end, hours):
  """
  Input: start and end as arrow.arrow.Arrow or datetime.date objects,
    hours as a float.

  Output: a float in the interval [0, 1], which is the percentage of full time.
  """
  days = (end - start).total_seconds()/60/60/24
  return (hours / 1600) * (365 / days)

def main():
  COURSES = {
    "DD1310": 
    "https://cloud.timeedit.net/kth/web/public01/ri.ics?sid=7&p=0.w%2C12.n&objects=453080.10&e=220609&enol=t&ku=29&k=1B9F3AD696BCA5C434C68950EFD376DD",
    "DD1317": 
    "https://cloud.timeedit.net/kth/web/public01/ri.ics?sid=7&p=0.w%2C12.n&objects=455995.10&e=220609&enol=t&ku=29&k=BA4400E3C003685549BC65AD9EAD3DC58E"
  }

  for course, url in COURSES.items():
    generate_signup_sheet(course, url)

if __name__ == "__main__":
    main()



"""Generate timesheets for TAs"""

import datetime
import io
import pkgutil
import os.path
import PIL.Image
from openpyxl import Workbook
from openpyxl.styles import Font
from openpyxl.styles import PatternFill
from openpyxl.drawing.image import Image
from openpyxl.styles import Alignment

def test():
    """Tests the module"""
    personnr = "123456-7890"
    name = "Alexander Baltatzis"
    email = "alba@kth.se"
    hourly_salary = 150
    course_leader = ('Daniel Bosk', 'dbosk@kth.se')
    HoD = "Karl Meinke"
    events = []
    events.append({"datum":"2022-12-03",
                  'tid':"8-10",
                  'kurskod':'DD1321',
                  'typ':"handl",
                  'timmar':2,
                  'koeff':1.33,
                  'omr_tid':2*1.33,
                  'belopp':hourly_salary*2*1.33})
    events.append({"datum":"2022-12-04",
                  'tid':"8-10",
                  'kurskod':'DD1321',
                  'typ':"övning",
                  'timmar':2,
                  'koeff':3,
                  'omr_tid':2*3,
                  'belopp':hourly_salary*2*3})
    events.append({"datum":"2022-12-05",
                  'tid':"8-10",
                  'kurskod':'DD1321',
                  'typ':"handl",
                  'timmar':2.1,
                  'koeff':1.33,
                  'omr_tid':2.1*1.33,
                  'belopp':hourly_salary*2.1*1.33})
    events.append({"datum":"2022-12-05",
                  'tid':"10-12",
                  'kurskod':'DD1310',
                  'typ':"handl",
                  'timmar':2,
                  'koeff':1.33,
                  'omr_tid':2*1.33,
                  'belopp':hourly_salary*2*1.33})

    make_xlsx(personnr, name, email, events,
              course_leader, HoD,
              course_leader_signature="signature.png")

def fit_image(img, height=None, width=None):
    """
    Rescales `img` to fit either `height` or `width` pixels --- not both, we 
    keep aspect ratio.

    `img` can be a string (path to file) or an image (PIL.Image).
    """
    img = Image(img)

    if height:
        scale = height/img.height
    elif width:
        scale = width/img.width
    else:
        raise KeyError("fit_image: neither height, nor width given.")

    img.height *= scale
    img.width *= scale
    return img

def make_xlsx(personnummer, name, email, events,
              course_leader, HoD,
              org = "JH", project = "1102",
              hourly_salary = 150,
              output = None,
              course_leader_signature = None,
              HoD_signature = None,
              logo = "kth.png"):
    """
    Generates a time report for a TA:
    - `name` and `email` are name and email for the TA.
    - `events` is a list of dictionaries, each dictionary containing:
        - datum: str
        - tid: str
        - kurskod: str
        - typ: str
        - timmar: float, koeff: float, omr_tid: float
        - belopp: float
    - `course_leader` is a tuple (name: str, email: str)
    - `HoD` the name of the Head of Department
    - `org`, `project` is the organization and project, both strings.
    - `hourly_salary` is the hourly salary, float.
    - `output` is the desired output (xlsx) filename.
    - `course_leader_signature` is an image (or path to one) containing the 
      course responsible's signature.
    - `HoD_signature` is an image (or path to one) containing the Head of 
      Department's signature.
    - `logo` is an image (or path to one) containing the logotype of the 
      university.
    """
    login = email.replace("@kth.se", "")
    if not output:
        output = login + "_tid_" + datetime.date.today().strftime("%Y-%m-%d.xlsx")

    wb = Workbook()
    ark = wb.active


    #############################################################
    # Logo
    ark.title = login + " " + datetime.date.today().strftime("%Y-%b")
    if isinstance(logo, str):
        logo = os.path.expanduser(logo)
    try:
        logo = fit_image(logo, height=80)
    except FileNotFoundError:
        img_data = PIL.Image.open(
                io.BytesIO(pkgutil.get_data(__name__, "kth.png")))
        logo = fit_image(img_data, height=80)
        ark.add_image(logo, "A1")
    else:
        ark.add_image(logo, "A1")

    #############################################################
    # kolumnstorlekar
    ark.column_dimensions['A'].width = 15  # 'Schemalagd tid'
    ark.column_dimensions['B'].width = 9   # 'Typ'           
    ark.column_dimensions['C'].width = 7   # 'timmar'      
    ark.column_dimensions['D'].width = 6   # 'koeff'         
    ark.column_dimensions['E'].width = 12  # 'omräknad tid'        
    ark.column_dimensions['F'].width = 6   # 'Timlön'     
    ark.column_dimensions['G'].width = 9   # 'Belopp'        
    ark.column_dimensions['H'].width = 9

    #############################################################
    # Börja på rad 6
    rad = "6"
    ark['A' + rad] = "Redovisning av arbetade timmar, time sheet"

    rad = incr(rad, 2)
    ark['A' + rad] = "Namn"
    ark['B' + rad] = name
    ark['B' + rad].fill = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    ark['C' + rad].fill = PatternFill(start_color="00EEECE1", 
                                      end_color="00EEECE1", fill_type="solid")

    ark['E' + rad] = 'Personnr'
    ark['F' + rad] = personnummer
    ark['F' + rad].fill = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    ark['G' + rad].fill = PatternFill(start_color="00EEECE1", 
                                      end_color="00EEECE1", fill_type="solid")

    rad = incr(rad)
    ark['A' + rad] = 'E-post'
    ark['B' + rad] = email
    ark['B' + rad].fill = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    ark['C' + rad].fill = PatternFill(start_color="00EEECE1", 
                                      end_color="00EEECE1", fill_type="solid")

    rad = incr(rad, 2)
    ark['A' + rad] = 'Kurskod'
    ark['B' + rad] = ''
    kurskoder = []
    for kol in events:
        if 'kurskod' in kol and kol['kurskod'] not in kurskoder:
            ark['B' + rad].value += kol['kurskod'] + " "
            kurskoder.append( kol['kurskod'] )
    ark['B' + rad].fill = PatternFill(start_color="00EEECE1", 
                                      end_color="00EEECE1", fill_type="solid")
    ark['C' + rad].fill = PatternFill(start_color="00EEECE1", 
                                      end_color="00EEECE1", fill_type="solid")

    rad = incr(rad, 2)
    ark['A' + rad] = 'Timmar ska anges inklusive förberedelsetid enligt schablon'
    rad = incr(rad)
    ark['A' + rad] = 'Ange typ av undervisning övning, handledning, möte, etc'

    rad = incr(rad, 2)
    ark['A' + rad] = 'Datum'
    ark['B' + rad] = 'Typ'
    ark['C' + rad] = 'Klocktimmar'
    ark['D' + rad] = 'koeff'
    ark['E' + rad] = 'Lönetimmar'
    ark['F' + rad] = 'Timlön'
    ark['G' + rad] = 'Belopp'

    for kol in ['C', 'D', 'E', 'F', 'G']:
        ark[kol+rad].alignment = Alignment(horizontal="right")


    #############################################################
    # Summering på sista raden 
    rad = incr(rad)   
    sist = incr(rad, len(events))
    ark['E'+sist].font = Font(bold=True)  
    ark['G'+sist].font = Font(bold=True)  
    ark['E'+sist].fill = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    #ark['G'+sist].fill = PatternFill(start_color="00C0C0C0", end_color="00C0C0C0", fill_type="solid")

    #############################################################
    # Matris med timredovisningen
    for i, kol in enumerate(events):
        ark['A'+rad].value = kol['datum']
        ark['B'+rad] = kol['typ']
        ark['C'+rad] = float(kol['timmar'])
        ark['D'+rad] = float(kol['koeff'])
        ark['E'+rad].value = float(kol['omr_tid'])
        ark['F'+rad].value = float(hourly_salary)
        ark['G'+rad].value = float(kol['belopp'])

        if i % 2 == 0:
            for kol in ['A', 'B', 'C', 'D', 'E', 'F', 'G']:
                ark[kol+rad].fill = PatternFill(start_color="00E0E0E0", end_color="00E0E0E0", fill_type="solid")
        
        if i == 0:
            tidsumma = "=SUM(E"+rad
            ark['G'+sist].value = '=G'+rad
        else:
            tidsumma += ',E'+rad 
            ark['G'+sist].value += '+G'+rad

        rad = incr(rad)

    ark['E'+sist].value = tidsumma + ')'

            
    #############################################################
    # Kontering
    rad = incr(sist, 3)
    ark['A'+rad].value = "Kontering"
    rad = incr(rad)
    ark['A'+rad].value = "Org.enhet"
    ark['A'+rad].fill  = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    ark['B'+rad].value = "Projekt"
    ark['B'+rad].fill = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    rad = incr(rad)
    ark['A'+rad].value = org
    ark['A'+rad].fill  = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")
    ark['B'+rad].value = project
    ark['B'+rad].fill  = PatternFill(start_color="00EEECE1", end_color="00EEECE1", fill_type="solid")

    #############################################################
    # Underskrift
    rad = incr(rad, 4)

    ark['A'+rad].value = "__________________________________"
    ark['E'+rad].value = "__________________________________"

    if isinstance(HoD_signature, str):
        HoD_signature = os.path.expanduser(HoD_signature)

    if HoD_signature:
        HoD_signature = fit_image(HoD_signature, height=60)
        ark.add_image(HoD_signature, "A"+incr(rad, -2))

    if isinstance(course_leader_signature, str):
        course_leader_signature = os.path.expanduser(course_leader_signature)

    if course_leader_signature:
        course_leader_signature = fit_image(course_leader_signature, height=60)
        ark.add_image(course_leader_signature, "E"+incr(rad, -2))

    rad = incr(rad)

    ark['A'+rad].value = "Ekonomisk attest " + HoD
    ark['E'+rad].value = "Kursansvarig"

    rad = incr(rad)
    ark['E'+rad].value = course_leader[0]
    rad = incr(rad)
    ark['E'+rad].value = course_leader[1]

    rad = incr(rad, 2)
    ark['A'+rad].value = "Underskriven blankett lämnas till HR"

    wb.save(output)

def incr(rad, i=1):
    """
    Bumps the row of a cell. Takes a string containing a number,
    increase by `i`.
    """
    return str(int(rad)+i)


#################################################################
#
# main - test
#
if __name__ == "__main__":
    test()




import datetime
from enum import Enum
import ics.icalendar
import logging
import pathlib
import sys
import typer
from typing_extensions import Annotated

from nytid.cli import courses as coursescli
from nytid import courses as courseutils
from nytid import schedules as schedutils
from nytid.signup import hr
from nytid.signup import sheets

from nytid.signup import sheets
from nytid.signup import utils
import os, platform, subprocess
from nytid.signup import sheets
from nytid.signup import utils
from nytid.signup import hr

SIGNUPSHEET_URL_PATH = "signupsheet.url"

cli = typer.Typer(name="signupsheets", help="Manage sign-up sheets for teaching")


def to_hours(td):
    return td.total_seconds() / 60 / 60


outpath_opt = typer.Option(
    help="Path where to write the sign-up sheet "
    "files. Default is in each course's "
    "data path."
)
edit_opt = typer.Option(help="If specified, opens each generated sheet " "for editing.")
url_arg = typer.Argument(
    help="The URL for the sign-up sheet. "
    "For Google Sheets, it's the same URL as the "
    "one shared with TAs to sign up. "
    "For others, it's a URL to a CSV file."
)


@cli.command(name="generate")
def generate_signup_sheet(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    outpath: Annotated[pathlib.Path, outpath_opt] = None,
    edit: Annotated[bool, edit_opt] = False,
):
    """
    Generates a sign-up sheet to be used with Google Sheets or similar for TAs to
    sign up.
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)

    now = datetime.datetime.now()
    timestamp = now.strftime("%Y%m%d-%H%M")
    for (course, register), course_conf in courses.items():
        num_students = course_conf.get("num_students")
        num_groups = course_conf.get("num_groups")

        group_size = round(num_students / num_groups)

        def needed_TAs(event):
            return utils.needed_TAs(event, group_size=group_size)

        if not outpath:
            data_root = courseutils.get_course_data(course, register)
            outfile = data_root.path / f"signup-{course}-{timestamp}.csv"
        else:
            outfile = outpath / f"signup-{course}-{timestamp}.csv"

        url = course_conf.get("ics")
        sheets.generate_signup_sheet(outfile, url, needed_TAs)
        if edit:
            the_os = platform.system()
            if the_os == "Darwin":
                subprocess.call(["open", outfile])
            elif the_os == "Windows":
                os.startfile(outfile)
            else:
                subprocess.call(["xdg-open", outfile])


@cli.command()
def set_url(
    course: Annotated[str, coursescli.course_arg_regex],
    url: Annotated[str, url_arg],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Sets the URL of the sign-up sheet for the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)
    for _, conf in courses.items():
        conf.set(SIGNUPSHEET_URL_PATH, url)


@cli.command()
def time(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Summarizes the time spent on teaching the course(s).
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    h_per_student = hr.hours_per_student(booked)

    for event, hours in h_per_student.items():
        print(f"{event}: {to_hours(hours):.2f} h/student")

    print(
        f"Booked: {to_hours(hr.total_hours(booked)):.2f} h "
        f"({to_hours(hr.max_hours(booked)):.2f} h)\n"
    )

    print("# Amanuenser")

    amanuensis = hr.compute_amanuensis_data(booked)

    for user, data in amanuensis.items():
        if not user:
            continue
        print(
            f"{user}: {data[2]:.2f} h, "
            f"{100*hr.compute_percentage(*data):.1f}%: "
            f"{data[0].format('YYYY-MM-DD')}--{data[1].format('YYYY-MM-DD')}"
        )

    print()
    print("# Hourly")

    for user, hours in hr.hours_per_TA(booked).items():
        if not user or user in amanuensis:
            continue
        print(f"{user}: {to_hours(hours):.2f} h")


@cli.command()
def users(
    course: Annotated[str, coursescli.course_arg_regex],
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
):
    """
    Prints the list of all usernames booked on the course.
    """
    registers = coursescli.registers_regex(register)
    courses = {
        course_reg: courseutils.get_course_config(*course_reg)
        for course_reg in coursescli.courses_regex(course, registers)
    }
    if not courses:
        sys.exit(1)

    booked = []
    for (course, register), config in courses.items():
        url = config.get(SIGNUPSHEET_URL_PATH)
        if "docs.google.com" in url:
            url = sheets.google_sheet_to_csv_url(url)
        booked += sheets.read_signup_sheet_from_url(url)

    for user in hr.hours_per_TA(booked):
        print(user)



from config import COURSES, SIGNUP
import arrow
import csv
from nytid.signup import hr
from nytid.signup import sheets
import operator
import sys

def to_hours(td):
    return td.total_seconds()/60/60

def shift_dates_forward(start_date, end_date):
    """Takes dates and shifts them so that start_date is tomorrow."""
    now = arrow.now()
    today = arrow.Arrow(now.year, now.month, now.day,
                        hour=start_date.hour, minute=start_date.minute,
                        second=start_date.second)

    if start_date > today:
        return start_date, end_date

    diff = (today-start_date).days
    return start_date.shift(days=diff+1), end_date.shift(days=diff+1)


if len(sys.argv) < 2:
    print(f"{sys.argv[0]}: requires argument 'username'",
          file=sys.stderr)
    print(f"{sys.argv[0]} <username> [<start date>]")
    sys.exit(1)
else:
    user = sys.argv[1]

if len(sys.argv) > 2:
    date = arrow.get(sys.argv[2])
else:
    date = None

booked = []

for course, url in SIGNUP.items():
    booked += sheets.read_signup_sheet_from_url(url)

amanuensis = hr.compute_amanuensis_data(booked,
                                        begin_date=date)
data = amanuensis[user]

#start, end = shift_dates_forward(data[0], data[1])
start = data[0]
end = data[1]

print(f"{user}: {data[2]:.2f} h, "
      f"{round(100*hr.compute_percentage(*data))}%: "
      f"{start.format('YYYY-MM-DD')}--{end.format('YYYY-MM-DD')}")

events = sheets.filter_events_by_TA(user, booked)
events = filter(lambda x: user in sheets.get_booked_TAs_from_csv(x)[0], booked)
events = list(map(lambda x: x[0:len(sheets.SIGNUP_SHEET_HEADER)] + [user], 
                  events))

for event, hours in hr.hours_per_event(events).items():
    print(f"{event}: {to_hours(hours)}")

print()

csvout = csv.writer(sys.stdout)

for event in events:
    csvout.writerow(event)
